/* Median-cut color reduction
 * (c) 1998 Jari Komppa aka Sol/Trauma
 *
 * Unauthorized use prohibited
 *
 * Request authorization for your use from Jari Komppa from
 * solar@compart.fi or jari.kompa@viherjuuri.fi, or search the
 * web for me if neither work.
 *
 *
 * Usage example:
 *   QUANT * q;
 *   int * idxmap;
 *   int idx[3];
 *   char * my_final_palette;
 *   q=newquant();
 *   idx[0]=qaddcolormap(q,my_palette,256);
 *   idx[1]=qaddcolormap(q,my_other_palette,256);
 *   idx[2]=qaddcolormap(q,my_yet_another_palette,256);
 *   qreduce(q,&idxmap,&my_final_palette,256);
 *   final_palette_index=*(idxmap+idx[1]+old_color_index_in_my_other_palette);
 */
#include <stdlib.h> /* exit, malloc, free */
#include <string.h> /* memset */

/* re-sorting after quantization */
#define RE_SORT
/* cut by light as well as r,g and b */
//#define CUT_LIGHTBASE
/* kill duplicates */
#define DUPENUKE

#define SWAP(a,b,c) {c=a; a=b; b=c; }

#pragma pack(0)

union colorchunk {
  char component[4]; // r,g,b,light
  int block;
};

struct colorstruc {
  struct colorstruc * next;
  int coloridx;
  union colorchunk data;
};

struct colormapstruc {
  struct colormapstruc * next;
  struct colorstruc * col;
};

typedef struct quantstruc {
  struct colorstruc * first;
  struct colorstruc * last;
  struct colorstruc * zeromap;
  struct colorstruc * zerolast;
  struct colormapstruc * colmap;
  struct colormapstruc * lastcolmap;
  int colors;
  int zeros;
} QUANT;

#pragma pack();

/*
 * local qcalloc
 * - Allocate memory; if failed, exit the program completely; if
 *   successful, clear the allocated memory to zero.
 */
static void * qcalloc(int size)
{
  void * temp;
  temp=malloc(size);
  if (temp==NULL) {
    exit(1);
  }
  memset(temp,0,size);
  return temp;
}

/*
 * public newquant
 * - initialize quantize structure
 */
QUANT * newquant(void)
{
  return (QUANT*) qcalloc(sizeof(QUANT));
}

/*
 * local add_zero
 * - add a color to the zero-list (black color)
 */
static void add_zero(QUANT * q, struct colorstruc * temp)
{
  temp->coloridx=q->colors;
  temp->data.block=-1;
  temp->next=NULL;
  if (q->zeromap==NULL) {
    q->zeromap=temp;
  } else {
    q->zerolast->next=temp;
  }
  q->zerolast=temp;
  q->colors++;
  q->zeros++;
}

/*
 * local add_color
 * - add a color to the color list
 */
static void add_color(QUANT * q, struct colorstruc * temp, int r, int g, int b)
{
  temp->coloridx=q->colors;
  temp->next=NULL;
  temp->data.component[0]=r;
  temp->data.component[1]=g;
  temp->data.component[2]=b;
  temp->data.component[3]=(r*3+g*6+b)/10;
  if (q->first==NULL) {
    q->first=temp;
  } else {
    q->last->next=temp;
  }
  q->last=temp;
  q->colors++;
}

/*
 * public qaddcolormap
 * - add a color map to be quantized
 *   returns index to the idxmap (see qreduce)
 */
int qaddcolormap(QUANT * q, char * colmap, int colors)
{
  int a,ret;
  struct colormapstruc * temp;
  ret=q->colors;
  temp=qcalloc(sizeof(struct colormapstruc));
  temp->col=qcalloc(sizeof(struct colorstruc)*colors);
  temp->next=NULL;
  if (q->colmap==NULL) {
    q->colmap=temp;
  } else {
    q->lastcolmap->next=temp;
  }
  q->lastcolmap=temp;
  for (a=0;a<colors;a++) {
    if ((*(colmap+a*3+0)==0) && (*(colmap+a*3+1)==0) && (*(colmap+a*3+2)==0)) {
      add_zero(q,temp->col+a);
    } else {
      add_color(q,temp->col+a,*(colmap+a*3+0),*(colmap+a*3+1),*(colmap+a*3+2));
    }
  }
  return ret;
}

/*
 * public qaddcolormap64
 * - add a color map to be quantized, all colors multiplied by 4
 *   (thus 64-based map will become 256-based map)
 *   returns index to the idxmap (see qreduce)
 */
int qaddcolormap64(QUANT * q, char * colmap, int colors)
{
  int a,ret;
  struct colormapstruc * temp;
  temp=qcalloc(sizeof(struct colormapstruc));
  temp->col=qcalloc(sizeof(struct colorstruc)*colors);
  temp->next=NULL;
  if (q->colmap==NULL) {
    q->colmap=temp;
  } else {
    q->lastcolmap->next=temp;
  }
  q->lastcolmap=temp;
  ret=q->colors;
  for (a=0;a<colors;a++) {
    if ((*(colmap+a*3+0)+*(colmap+a*3+1)+*(colmap+a*3+2))==0) {
      add_zero(q,temp->col+a);
    } else {
      add_color(q,temp->col+a,*(colmap+a*3+0)*4,*(colmap+a*3+1)*4,*(colmap+a*3+2)*4);
    }
  }
  return ret;
}

/*
 * local free_colmaps
 * - frees all allocated colormaps
 */
static void free_colmaps(QUANT * q)
{
  struct colormapstruc * temp, * temp2;
  temp=q->colmap;
  while (temp!=NULL) {
    temp2=temp;
    free(temp->col);
    temp=temp->next;
    free(temp2);
  }
}

/*
 * local examine_group
 * - Find out the largest component in sub-colorspace and return it and
 *   its size.
 */
static void examine_group(struct colorstruc * g,int * groupcomponent, int * size)
{
  int rmin,rmax,gmin,gmax,bmin,bmax,rs,gs,bs,a,groupsize;
#ifdef CUT_LIGHTBASE
  int lmin,lmax,ls;
  lmin=lmax=g->data.component[3];
  ls=0;
#endif
  rmin=rmax=g->data.component[0];
  gmin=gmax=g->data.component[1];
  bmin=bmax=g->data.component[2];
  rs=gs=bs=0;
  g=g->next;
  while (g!=NULL) {
    a=g->data.component[0];
    if (rmin>a) rmin=a; else
    if (rmax<a) rmax=a;
    rs+=a;
    a=g->data.component[1];
    if (gmin>a) gmin=a; else
    if (gmax<a) gmax=a;
    gs+=a;
    a=g->data.component[2];
    if (bmin>a) bmin=a; else
    if (bmax<a) bmax=a;
    bs+=a;
#ifdef CUT_LIGHTBASE
    a=g->data.component[3];
    if (lmin>a) lmin=a; else
    if (lmax<a) lmax=a;
    ls+=a;
#endif
    g=g->next;
  }
  groupsize=rmax-rmin;
  *groupcomponent=0;
  *size=rs;
  if ((gmax-gmin)>groupsize) {
    groupsize=gmax-gmin;
    *groupcomponent=1;
    *size=gs;
  }
#ifdef CUT_LIGHTBASE
  if ((lmax-lmin)>groupsize) {
    groupsize=lmax-lmin;
    *groupcomponent=3;
    *size=ls;
  }
#endif
  if ((bmax-bmin)>groupsize) {
    *groupcomponent=2;
    *size=bs;
  }
}

/*
 * local cut_group
 * - cut a sub-colorspace in two sub-colorspaces by splitting the
 *   linked list at median. Returns the new sub-colorspace.
 */
static struct colorstruc * cut_group(struct colorstruc * g, int component, int max)
{
  int median, count;
  struct colorstruc * second=NULL, * fore=NULL;
  median=max/2;
  count=0;
  while (count<median) {
    count+=g->data.component[component];
    fore=second;
    second=g;
    g=g->next;
  }
  if (fore==NULL) {
    second->next=NULL;
    second=g;
  } else {
    fore->next=NULL;
  }
  return second;
}

/*
 * local sort_group
 * - Sorts a sub-colorspace by given component with one-pass radix sort.
 */
static struct colorstruc * sort_group(struct colorstruc * col,int component)
{
  struct colorstruc * bucket[256], * top[256];
  int a,s,l;
  memset(bucket,0,sizeof(bucket)); /* set bucket[n] to NULL */
/* Step 1: cut list into N lists */
  while (col!=NULL) {
    a=col->data.component[component];
    if (bucket[a]==NULL) {
      bucket[a]=col;
    } else {
      top[a]->next=col;
    }
    top[a]=col;
    col=col->next;
    top[a]->next=NULL;
  }
/* Step 2: re-link the list. */
  /* s = index of the first full bucket
   * l = last used bucket
   * a = counter
   */
  s=0;
  while (bucket[s]==NULL) s++;
  a=l=s;
  a++;
  while (a<256) {
    if (bucket[a]!=NULL) {
      top[l]->next=bucket[a];
      l=a;
    }
    a++;
  }
  return(bucket[s]);
}

/*
 * local dupenuke
 * - kill duplicate colors by first sorting the list by all components,
 *   then comparing adjacent colors, moving duplicates to the zero-list.
 */
static void dupenuke(QUANT *q) {
  struct colorstruc * col, * last;
  int lastidx, lastcol;
  q->first=sort_group(q->first,0);
  q->first=sort_group(q->first,1);
  q->first=sort_group(q->first,2);
  last=col=q->first;
  lastidx=col->coloridx;
  lastcol=col->data.block;
  col=col->next;
  while (col!=NULL) {
    if (col->data.block==lastcol) { // duplicate
      col->data.block=lastidx; // set datablock to point to real color instead
      last->next=col->next;    // removed from the list
      if (q->zeromap==NULL) {
        q->zeromap=col;
      } else {
        q->zerolast->next=col;
      }
      q->zerolast=col;
      q->zerolast->next=NULL;  // added to zero-list
      col=last;
      q->zeros++;
    } else { // not dupe
      lastidx=col->coloridx;
      lastcol=col->data.block;
    }
    last=col;
    col=col->next;
  }
}

/*
 * public qreduce
 * - do the quantize, build index map, free all allocated resources.
 *   returns total number of colors in outputted map.
 */
int qreduce(QUANT * q, int ** idxmap, char ** pal, int palwid)
{
  int i, n, totalcolors, comp[3], count, groups;
  struct colorstruc * col;
  struct colorstruc ** group;
  int * groupcomponent, * groupcomponentsize, * groupsorted;
#ifdef RE_SORT
  int * remapmap;
  char * temppal,*tempcp;
  QUANT * reorder;
#endif
  totalcolors=q->colors;
  *pal=qcalloc(sizeof(char)*palwid*3);
  *idxmap=qcalloc(sizeof(int)*totalcolors);
  group=qcalloc(sizeof(struct colorstruc *)*palwid);
  groupsorted=qcalloc(sizeof(int)*palwid);
  groupcomponent=qcalloc(sizeof(int)*palwid);
  groupcomponentsize=qcalloc(sizeof(int)*palwid);
#ifdef DUPENUKE
  dupenuke(q);
#endif
  if ((q->colors-q->zeros)<palwid) {
    /*
     * If number of input colors is less than requested output,
     * just sort the colors, make sure color 0 is black, and
     * so on. Don't do any real reducing, that is.
     */
    q->first=sort_group(q->first,3);
    i=1;
    *(*pal+0)=0;
    *(*pal+1)=0;
    *(*pal+2)=0;
    col=q->first;
    while (col!=NULL) {
      *(*idxmap+col->coloridx)=i;
      *(*pal+i*3+0)=col->data.component[0];
      *(*pal+i*3+1)=col->data.component[1];
      *(*pal+i*3+2)=col->data.component[2];
      col=col->next;
      i++;
    }
    col=q->zeromap;
    while (col!=NULL) {
      if (col->data.block==-1)
        *(*idxmap+col->coloridx)=0;
      else
        *(*idxmap+col->coloridx)=*(*idxmap+col->data.block);
      col=col->next;
    }
    free(group);
    free(groupsorted);
    free(groupcomponent);
    free(groupcomponentsize);
    free_colmaps(q);
    free(q);
    return totalcolors;
  }
  groups=1;
  group[0]=q->first;
  examine_group(group[0],&groupcomponent[0],&groupcomponentsize[0]);
#ifdef DUPENUKE
  groupsorted[0]=2;
#else
  groupsorted[0]=-1;
#endif
  while (groups<(palwid-1)) {
    i=0;
    for (n=0;n<groups;n++)
      if (groupcomponentsize[n]>groupcomponentsize[i]) i=n;
    if (groupsorted[i]!=groupcomponent[i]) {
      group[i]=sort_group(group[i],groupcomponent[i]);
      groupsorted[i]=groupcomponent[i];
    }
    group[groups]=cut_group(group[i],groupcomponent[i],groupcomponentsize[i]);
    if (group[groups]==NULL) break;
    examine_group(group[i],&groupcomponent[i],&groupcomponentsize[i]);
    examine_group(group[groups],&groupcomponent[groups],&groupcomponentsize[groups]);
    groupsorted[groups]=groupsorted[i];
    groups++;
  }
  *(*pal+0)=0;
  *(*pal+1)=0;
  *(*pal+2)=0;
  for (i=0;i<groups;i++) {
    col=group[i];
    count=comp[0]=comp[1]=comp[2]=0;
    while (col!=NULL) {
      count++;
      *(*idxmap+col->coloridx)=i+1;
      comp[0]+=col->data.component[0];
      comp[1]+=col->data.component[1];
      comp[2]+=col->data.component[2];
      col=col->next;
    }
    *(*pal+i*3+3)=comp[0]/count;
    *(*pal+i*3+4)=comp[1]/count;
    *(*pal+i*3+5)=comp[2]/count;
  }
  col=q->zeromap;
  while (col!=NULL) {
    if (col->data.block==-1)
      *(*idxmap+col->coloridx)=0;
    else
      *(*idxmap+col->coloridx)=*(*idxmap+col->data.block);
    col=col->next;
  }
  free(group);
  free(groupcomponent);
  free(groupcomponentsize);
  free_colmaps(q);
  free(q);
#ifdef RE_SORT
  reorder=newquant();
  qaddcolormap(reorder,*pal,palwid);
  qreduce(reorder,&remapmap,&temppal,palwid);
  SWAP(*pal,temppal,tempcp);
  free(temppal);
  for (i=0;i<totalcolors;i++)
    *(*idxmap+i)=*(remapmap+*(*idxmap+i));
  free(remapmap);
#endif /* RE_SORT */
  return totalcolors; /* total width of idxmap */
}
